import os

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django_extensions.db.models import TimeStampedModel
from wagtail.admin.panels import FieldPanel, MultiFieldPanel
from wagtail.api.v2.utils import get_full_url
from wagtail.fields import StreamField
from wagtail.images.blocks import ImageChooserBlock
from wagtail.images.models import Image
from wagtail_modeladmin.helpers import AdminURLHelper

from geomanager.blocks import (
    FileLayerPointAnalysisBlock,
    FileLayerAreaAnalysisBlock, InlineLegendBlock
)
from geomanager.helpers import get_raster_layer_files_url
from geomanager.models.raster_style import RasterStyle
from geomanager.models.core import Dataset, BaseLayer
from geomanager.settings import geomanager_settings
from geomanager.storage import OverwriteStorage
from geomanager.utils import DATE_FORMAT_CHOICES
from geomanager.validators import validate_directory_name


class RasterFileLayer(TimeStampedModel, BaseLayer):
    dataset = models.ForeignKey(Dataset, on_delete=models.CASCADE, related_name="raster_file_layers",
                                verbose_name=_("dataset"))
    date_format = models.CharField(max_length=100, choices=DATE_FORMAT_CHOICES, blank=True, null=True,
                                   default="yyyy-MM-dd HH:mm",
                                   verbose_name=_("Display Format for DateTime Selector"))
    style = models.ForeignKey("RasterStyle", null=True, blank=True, on_delete=models.SET_NULL, verbose_name=_("style"))
    use_custom_legend = models.BooleanField(default=False, verbose_name=_("Use custom legend"))
    legend = StreamField([
        ('legend', InlineLegendBlock(label=_("Custom Legend")),),
        ('legend_image', ImageChooserBlock(label=_("Custom Image")),),
    ], use_json_field=True, null=True, blank=True, max_num=1, verbose_name=_("Legend"), )

    auto_ingest_from_directory = models.BooleanField(default=False, verbose_name=_("Auto ingest from directory"))
    auto_ingest_use_custom_directory_name = models.BooleanField(default=False,
                                                                verbose_name=_("Use custom directory name"))
    auto_ingest_custom_directory_name = models.CharField(max_length=255, blank=True, null=True, unique=True,
                                                         validators=[validate_directory_name],
                                                         verbose_name=_("Custom directory name"))
    auto_ingest_nc_data_variable = models.CharField(max_length=100, blank=True, null=True,
                                                    verbose_name=_("Data variable for netCDF data auto ingest"),
                                                    help_text=_("The name of the data variable to use, "
                                                                "if ingesting from netCDF files"))

    analysis = StreamField([
        ('point_analysis', FileLayerPointAnalysisBlock(label=_("Point Analysis")),),
        ('area_analysis', FileLayerAreaAnalysisBlock(label=_("Area Analysis")),),
    ], block_counts={'point_analysis': {'max_num': 1}, 'area_analysis': {'max_num': 1}}, use_json_field=True,
        null=True, blank=True, max_num=2, verbose_name=_("Analysis"), )

    class Meta:
        verbose_name = _("Raster File Layer")
        verbose_name_plural = _("Raster File Layers")
        ordering = ['order']

    panels = [
        FieldPanel("dataset"),
        FieldPanel("title"),
        FieldPanel("default"),
        FieldPanel("date_format"),
        FieldPanel("style"),
        FieldPanel("use_custom_legend"),
        FieldPanel("legend"),
        FieldPanel("auto_ingest_from_directory"),

        MultiFieldPanel([
            FieldPanel("auto_ingest_use_custom_directory_name"),
            FieldPanel("auto_ingest_custom_directory_name", classname="show-if-custom-dir-name"),
            FieldPanel("auto_ingest_nc_data_variable"),
        ], heading=_("Auto ingest settings")),

        FieldPanel("analysis"),
    ]

    def __str__(self):
        return f"{self.dataset.title} - {self.title}"

    @property
    def raster_files_count(self):
        return self.raster_files.count()

    def get_uploads_list_url(self):
        url = get_raster_layer_files_url(self.pk)
        return url

    def get_style_url(self):
        url = {"action": _("Create Style")}
        style_admin_helper = AdminURLHelper(RasterStyle)
        if self.style:
            url.update({
                "action": _("Edit Style"),
                "url": style_admin_helper.get_action_url("edit", self.style.pk)
            })
        else:
            url.update({
                "url": style_admin_helper.get_action_url("create") + f"?layer_id={self.pk}"
            })
        return url

    @property
    def base_tile_url(self):
        base_tiles_url = reverse("raster_tiles", args=(self.id, 0, 0, 0))
        base_tiles_url = base_tiles_url.replace("/0/0/0", r"/{z}/{x}/{y}")
        return f"{base_tiles_url}"

    def get_tile_json_url(self, request=None):
        tile_json_url = reverse("raster_file_tile_json", args=(self.id,))
        if request:
            tile_json_url = get_full_url(request, tile_json_url)

        return tile_json_url

    def layer_config(self, request=None):
        base_tile_url = self.base_tile_url
        if request:
            base_tile_url = get_full_url(request, base_tile_url)

        tile_url = f"{base_tile_url}?time={{time}}"

        if self.dataset.can_clip:
            tile_url = tile_url + "&geostore_id={geostore_id}"

        layer_config = {
            "type": "raster",
            "source": {
                "type": "raster",
                "tiles": [tile_url]
            }
        }

        return layer_config

    @property
    def params(self):
        params = {
            "time": ""
        }

        if self.dataset.can_clip:
            params.update({"geostore_id": ""})

        return params

    @property
    def param_selector_config(self):
        if self.dataset.multi_layer:
            default_layer = self.dataset.layers.filter(default=True).exclude(pk=self.pk).first()
            if default_layer:
                return default_layer.param_selector_config

        time_config = {
            "key": "time",
            "required": True,
            "sentence": "{selector}",
            "type": "datetime",
            "availableDates": [],
        }

        if self.date_format:
            if self.date_format == "pentadal":
                time_config.update({
                    "dateFormat": {"currentTime": "MMM yyyy", "asPeriod": "pentadal"},
                })
            elif self.date_format == "dekadal":
                time_config.update({
                    "dateFormat": {"currentTime": "MMM yyyy", "asPeriod": "dekadal"},
                })
            else:
                time_config.update({
                    "dateFormat": {"currentTime": self.date_format},
                })
        else:
            time_config.update({
                "dateFormat": {"currentTime": "yyyy-MM-dd HH:mm"},
            })

        return [time_config]

    def get_legend_config(self, request=None):
        config = {
            "type": "choropleth",
            "items": []
        }

        if self.style:
            if self.use_custom_legend:
                legend_block = self.legend

                # only one legend block entry is expected
                if legend_block:
                    legend_block = legend_block[0]

                if legend_block:
                    if isinstance(legend_block.value, Image):
                        image_url = legend_block.value.file.url
                        image_url = get_full_url(request, image_url)
                        config.update({"type": "image", "imageUrl": image_url})

                        return config

                    data = legend_block.block.get_api_representation(legend_block.value)

                    config.update({"type": data.get("type")})

                    for item in data.get("items"):
                        config["items"].append({
                            "name": item.get("value"),
                            "color": item.get("color")
                        })

                    return config

            return self.style.get_legend_config()

        return config

    def get_analysis_config(self):
        analysis_config = {}

        for analysis in self.analysis:
            data = analysis.block.get_api_representation(analysis.value)

            if analysis.block_type == "point_analysis":
                unit = data.get("unit")
                if data.get("instance_data_enabled"):
                    analysis_config.update({
                        "pointInstanceAnalysis": {"unit": unit}
                    })
                if data.get("timeseries_data_enabled"):
                    analysis_config.update({
                        "pointTimeseriesAnalysis": {
                            "unit": unit,
                            "chartType": data.get("timeseries_chart_type"),
                            "chartColor": data.get("timeseries_chart_color")
                        }
                    })
            if analysis.block_type == "area_analysis":
                unit = data.get("unit")
                if data.get("instance_data_enabled"):
                    analysis_config.update({
                        "areaInstanceAnalysis": {
                            "unit": unit,
                            "valueType": data.get("instance_value_type")
                        }
                    })
                if data.get("timeseries_data_enabled"):
                    analysis_config.update({
                        "areaTimeseriesAnalysis": {
                            "unit": unit,
                            "aggregationMethod": data.get("timeseries_aggregation_method"),
                            "chartType": data.get("timeseries_chart_type"),
                            "chartColor": data.get("timeseries_chart_color")
                        }
                    })

        return analysis_config

    def get_tile_json(self, request=None):
        base_tile_url = self.base_tile_url
        timestamps = list(self.raster_files.all().values_list("time", flat=True))

        if request:
            base_tile_url = get_full_url(request, base_tile_url)

        tile_json = {
            "tilejson": "3.0.0",
            "name": self.title,
            "scheme": "xyz",
            "tiles": [base_tile_url],
            "minzoom": 0,
            "maxzoom": 20,
            "time_parameter": "time",
            "timestamps": [t.strftime("%Y-%m-%dT%H:%M:%S.000Z") for t in timestamps]
        }

        return tile_json

    def clean(self):
        # if adding a layer to a dataset that already has a layer and is not multi layer
        if self._state.adding:
            if self.dataset.has_layers() and not self.dataset.multi_layer:
                raise ValidationError(_("Can not add layer because the dataset is not marked as Multi Layer. "
                                        "To add multiple layers to a dataset, please mark the dataset as "
                                        "Multi Layer and try again"))


@receiver(post_save, sender=RasterFileLayer)
def create_auto_ingest_directory(sender, instance, created, **kwargs):
    if instance.auto_ingest_from_directory:
        auto_ingest_raster_data_dir = geomanager_settings.get("auto_ingest_raster_data_dir")

        dir_name = str(instance.pk)

        if auto_ingest_raster_data_dir and os.path.isabs(auto_ingest_raster_data_dir):

            if instance.auto_ingest_use_custom_directory_name and instance.auto_ingest_custom_directory_name:
                dir_name = instance.auto_ingest_custom_directory_name

            if not os.path.exists(auto_ingest_raster_data_dir):
                os.makedirs(auto_ingest_raster_data_dir)

            directory_path = os.path.join(auto_ingest_raster_data_dir, dir_name)

            # Create the directory if it doesn't exist
            if not os.path.exists(directory_path):
                os.makedirs(directory_path)


def layer_raster_file_dir_path(instance, filename):
    file_dir = f"raster_files/{type(instance.layer).__name__}-{instance.layer.pk}/{filename}"
    return file_dir


class LayerRasterFile(TimeStampedModel):
    layer = models.ForeignKey(RasterFileLayer, on_delete=models.CASCADE, related_name="raster_files",
                              verbose_name=_("layer"))
    file = models.FileField(upload_to=layer_raster_file_dir_path, storage=OverwriteStorage, verbose_name=_("file"))
    time = models.DateTimeField(verbose_name=_("time"),
                                help_text=_("Time for the raster file. This can be the time the data was acquired, "
                                            "or the date and time for which the data applies", ))
    raster_metadata = models.JSONField(blank=True, null=True)

    class Meta:
        verbose_name = _("Layer Raster File")
        verbose_name_plural = _("Layer Raster Files")
        ordering = ["-time"]
        unique_together = ('layer', 'time')

    panels = [
        FieldPanel("time"),
        FieldPanel("raster_metadata", read_only=True),
    ]

    def __str__(self):
        return f"{self.layer.title} - {self.time}"

    @property
    def thumbnail_url(self):
        url = reverse("raster_file_thumbnail", kwargs={"file_id": self.pk})
        return url

    @property
    def time_str(self):
        return self.time.strftime("%Y-%m-%dT%H:%M:%S.000Z")


class RasterUpload(TimeStampedModel):
    dataset = models.ForeignKey(Dataset, blank=True, null=True, on_delete=models.SET_NULL, verbose_name=_("dataset"))
    file = models.FileField(upload_to="raster_uploads", verbose_name=_("file"))
    raster_metadata = models.JSONField(blank=True, null=True)

    class Meta:
        verbose_name = _("Raster Upload")
        verbose_name_plural = _("Raster Uploads")

    panels = [
        FieldPanel("layer"),
        FieldPanel("file"),
    ]

    def __str__(self):
        return f"{self.dataset} - {self.created}"



from django import forms
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from wagtail_modeladmin.views import CreateView

from geomanager.admin.base import BaseModelAdmin, ModelAdminCanHide
from geomanager.models import RasterFileLayer, RasterStyle


class RasterStyleCreateView(CreateView):
    def get_form(self):
        form = super().get_form()

        try:
            # check if we have layer_id in GET
            layer_id = self.request.GET.get("layer_id")

            # add hidden layer_id field to form. We will use it later to update the layer style
            if layer_id:
                layer = RasterFileLayer.objects.get(pk=layer_id)
                form.fields["layer_id"] = forms.CharField(required=False, widget=forms.HiddenInput())
                form.initial.update({"layer_id": layer.pk})
        except Exception:
            pass

        return form

    def form_valid(self, form):
        response = super().form_valid(form)

        try:
            # check if we have layer_id in data
            layer_id = form.data.get("layer_id")

            if layer_id:
                # assign this layer the just created style
                layer = RasterFileLayer.objects.get(pk=layer_id)
                layer.style = self.instance
                layer.save()
        except Exception:
            pass

        return response


class RasterStyleModelAdmin(BaseModelAdmin, ModelAdminCanHide):
    model = RasterStyle
    exclude_from_explorer = True
    create_view_class = RasterStyleCreateView
    list_display = ("__str__", "min", "max")
    form_view_extra_js = ["geomanager/js/raster_style_extra.js"]
    menu_icon = "palette"

    def __init__(self, parent=None):
        super().__init__(parent)
        self.list_display = (list(self.list_display) or []) + ["preview"]
        self.preview.__func__.short_description = _("Legend Preview")

    def get_index_view_extra_js(self):
        js = [
            "geomanager/js/vendor/d3.min.js",
            "geomanager/js/raster_style_index_extra.js",
        ]

        return js

    def preview(self, obj):
        legend_config = obj.get_legend_config()

        colors_str = ",".join([item.get("color") for item in legend_config.get("items")])
        context = {
            "legend_config": legend_config,
            "raster_style": obj,
            "colors_str": colors_str,
        }
        html = render_to_string("geomanager/modeladmin/raster_style_legend_preview.html", context=context)

        return mark_safe(html)



# Generated by Django 4.1.10 on 2023-11-19 17:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geomanager', '0033_alter_layerrasterfile_file_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='rastertilelayer',
            name='time_parameter_name',
            field=models.CharField(blank=True, default='time', help_text='Name of the time parameter in the url', max_length=100, null=True, verbose_name='Time Parameter Name'),
        ),
        migrations.AddField(
            model_name='vectortilelayer',
            name='time_parameter_name',
            field=models.CharField(blank=True, default='time', help_text='Name of the time parameter in the url', max_length=100, null=True, verbose_name='Time Parameter Name'),
        ),
    ]



# Generated by Django 4.1.10 on 2023-11-14 10:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geomanager', '0031_remove_geomanagersettings_cap_auto_refresh_interval_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='rasterfilelayer',
            name='auto_ingest_nc_data_variable',
            field=models.CharField(blank=True, max_length=100, null=True, verbose_name='Data variable for netCDF data auto ingest'),
        ),
    ]



# Generated by Django 4.1.10 on 2023-07-06 10:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geomanager', '0004_alter_dataset_can_clip'),
    ]

    operations = [
        migrations.DeleteModel(
            name='CountryBoundary',
        ),
        migrations.RemoveField(
            model_name='geomanagersettings',
            name='country',
        ),
        migrations.AddField(
            model_name='geostore',
            name='id3',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='geostore',
            name='name_3',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='geostore',
            name='name_4',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]



from functools import wraps

from django.utils.cache import patch_cache_control


def revalidate_cache(view_func):
    """
    Decorator that adds must_revalidate header to a response
    """

    @wraps(view_func)
    def _wrapped_view_func(request, *args, **kwargs):
        # Ensure argument looks like a request.
        if not hasattr(request, "META"):
            raise TypeError(
                "revalidate_cache didn't receive an HttpRequest. If you are "
                "decorating a classmethod, be sure to use @method_decorator."
            )
        response = view_func(request, *args, **kwargs)
        patch_cache_control(response, must_revalidate=True)
        return response

    return _wrapped_view_func



import glob
import json
import os
import tempfile
from subprocess import Popen, PIPE
from urllib.parse import unquote
from zipfile import ZipFile

import geopandas as gpd
from django.contrib.gis.db import models
from django.db import connection
from django.utils.translation import gettext as _

from geomanager.errors import (
    NoShpFound,
    NoShxFound,
    NoDbfFound,
    InvalidFile,
    InvalidGeomType,
    GeomValidationNotImplemented
)

POSTGRES_DATA_TYPES_DJANGO_FIELDS_MAPPING = {
    'smallint': models.SmallIntegerField,
    'integer': models.IntegerField,
    'bigint': models.BigIntegerField,
    'decimal': models.DecimalField,
    'numeric': models.DecimalField,
    'real': models.FloatField,
    'double precision': models.FloatField,
    'character varying': models.CharField,
    'character': models.CharField,
    'text': models.TextField,
    'date': models.DateField,
    'time': models.TimeField,
    'timestamp': models.DateTimeField,
    'timestamp with time zone': models.DateTimeField,
    'interval': models.DurationField,
    'boolean': models.BooleanField,
    "Point": models.PointField,
    "MultiPoint": models.MultiPointField,
}

GEOM_TYPES = {
    "POINT": "Point",
    "MULTIPOINT": "MultiPoint",
    "POLYGON": "Polygon",
    "MULTIPOLYGON": "MultiPolygon",
    "LINESTRING": "LineString",
    "MULTILINESTRING": "MultiLineString",
}


def ogr2pg(file_path, table_name, db_settings, srid=4326, overwrite=False):
    # construct db connection options from uri
    db_host = db_settings.get("host")
    db_port = db_settings.get("port")
    db_user = db_settings.get("user")
    db_password = db_settings.get("password")
    db_name = db_settings.get("name")
    # unquote incase encoded
    db_password = unquote(db_password)

    pg_service_schema = db_settings.get("pg_service_schema")

    # construct ogr2ogr command
    # notable option: -lco FID=gid  - Use custom fid column name. ogr by default gives ogc_fid. We use gid instead
    cmd = ["ogr2ogr", "-f", "PostgreSQL",
           f"PG:host={db_host} port={db_port} user={db_user} password={db_password} dbname={db_name}", file_path,
           "-nln", table_name, "-lco", f"FID=gid", "-lco", f"SCHEMA={pg_service_schema}", "-lco", "GEOMETRY_NAME=geom",
           "-nlt", "PROMOTE_TO_MULTI", "-lco", "PRECISION=NO"
           ]

    # overwrite existing
    if overwrite:
        cmd.append("-lco")
        cmd.append("OVERWRITE=YES")

    p1 = Popen(cmd, stdout=PIPE, stderr=PIPE)

    p1.stdout.close()

    stdout, stderr = p1.communicate()

    # TODO: Catch possible errors better
    if stderr:
        raise Exception(stderr)

    full_table_name = f"{pg_service_schema}.{table_name}"

    pg_table = {"table_name": full_table_name, "srid": srid}

    info = get_postgis_table_info(pg_service_schema, table_name)

    pg_table.update({**info})

    return pg_table


def extract_zipped_shapefile(shp_zip_path, out_dir):
    # unzip file
    with ZipFile(shp_zip_path, 'r') as zip_obj:
        for filename in zip_obj.namelist():
            # ignore __macosx files
            if not filename.startswith('__MACOSX/'):
                zip_obj.extract(filename, out_dir)

    # Use the first available shp
    shp = glob.glob(f"{out_dir}/*.shp") or glob.glob(f"{out_dir}/*/*.shp")

    if not shp:
        raise NoShpFound(_("No shapefile found in provided zip file"))

    shp_fn = os.path.splitext(shp[0])[0]
    shp_dir = os.path.dirname(shp_fn)

    files = [os.path.join(shp_dir, f) for f in os.listdir(shp_dir)]

    # check for .shx
    if f"{shp_fn}.shx" not in files:
        raise NoShxFound(_("No .shx file found in provided zip file"))

    # check for .dbf
    if f"{shp_fn}.dbf" not in files:
        raise NoDbfFound(_("No .dbf file found in provided zip file"))

    # return first shp path
    return shp[0]


def ogr_db_import(file_path, table_name, db_settings, overwrite=False, validate_geom_types=None):
    file_extension = os.path.splitext(file_path)[1]

    # handle shapefile
    if file_extension == ".zip":
        with tempfile.TemporaryDirectory() as tmpdir:
            shp = extract_zipped_shapefile(file_path, tmpdir)
            if validate_geom_types and isinstance(validate_geom_types, list):
                validate_vector_geom_type(shp, validate_geom_types, vector_format="shp")

            table_info = ogr2pg(shp, table_name, db_settings, overwrite=overwrite)

            return table_info

    # handle geojson
    if file_extension == ".geojson":
        table_info = ogr2pg(file_path, table_name, db_settings, overwrite=overwrite)
        return table_info

    # handle geopackage
    if file_extension == ".gpkg":
        table_info = ogr2pg(file_path, table_name, db_settings, overwrite=overwrite)
        return table_info

    raise InvalidFile(message=_('Unsupported file type'))


def get_postgis_table_info(schema, table_name):
    # Get all the column names with corresponding data types
    with connection.cursor() as cursor:
        columns_sql = f"""SELECT column_name, data_type 
        FROM information_schema.columns 
        WHERE table_name = '{table_name}' AND table_schema = '{schema}'
        """
        cursor.execute(columns_sql)
        results = cursor.fetchall()
        column_data_types = [{"name": row[0], "data_type": row[1]} for row in results if row[0] != "geom"]

        # Get the extents of all the data in the table
        extent_sql = f"SELECT ST_Extent(geom) FROM {schema}.{table_name}"
        cursor.execute(extent_sql)
        bbox_text = cursor.fetchone()[0]
        bbox_str = bbox_text.replace("BOX(", "").replace(")", "").replace(" ", ",").split(",")

        # Get the geometry type of the 'geom' column
        geom_sql = f"""SELECT GeometryType(geom)
        FROM {schema}.{table_name} LIMIT 1
        """
        cursor.execute(geom_sql)
        result = cursor.fetchone()[0]
        geometry_type = result.replace('ST_', '')

    return {
        'properties': column_data_types,
        'bounds': bbox_str,
        'geom_type': GEOM_TYPES[geometry_type],
    }


def drop_vector_table(table_name):
    with connection.cursor() as cursor:
        cursor.execute(f"DROP TABLE IF EXISTS {table_name};")


def create_feature_collection_from_geom(geom):
    return {
        "type": "FeatureCollection",
        "features": [
            {
                "type": "Feature",
                "geometry": json.loads(geom.geojson),
                "properties": {}
            }
        ]
    }


def validate_vector_geom_type(file_path, valid_geom_types, vector_format):
    if vector_format == "shp":
        gdf = gpd.read_file(file_path, layer=0)
        geom_types = gdf.geometry.geom_type.unique()

        for geom_type in geom_types:
            if geom_type not in valid_geom_types:
                error_message = _("Invalid geometry type. Expected one of %(valid_geom_types)s. Not %(geom_type)s") % {
                    "valid_geom_types": valid_geom_types, "geom_type": geom_type}
                raise InvalidGeomType(error_message)
    else:
        error_message = _("Geometry Type validation not implemented for format %(vector_format)s") % {
            "vector_format": vector_format}
        raise GeomValidationNotImplemented(error_message)


def get_model_field(pg_data_type):
    return POSTGRES_DATA_TYPES_DJANGO_FIELDS_MAPPING.get(pg_data_type)



# Generated by Django 4.2.10 on 2024-04-23 07:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geomanager', '0047_geomanagersettings_map_disclaimer_page_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='geomanagersettings',
            name='map_disclaimer_text',
            field=models.CharField(blank=True, default='The boundaries and names shown and the designations used on this map do not imply the expression of any opinion whatsoever concerning the legal status of any country, territory, city or area or of its authorities, or concerning the delimitation of its frontiers or boundaries', help_text='MapViewer Map disclaimer text. Maximum of 350 characters. Add more details in the Map Disclaimer Page and link it below', max_length=350, null=True, verbose_name='Map Disclaimer Text'),
        ),
    ]



from django.db import models
from django.utils.translation import gettext_lazy as _
from django_extensions.db.models import TimeStampedModel
from modelcluster.models import ClusterableModel
from wagtail.admin.panels import FieldPanel, MultiFieldPanel
from wagtail.api.v2.utils import get_full_url
from wagtail.fields import StreamField
from wagtail.images.blocks import ImageChooserBlock
from wagtail.images.models import Image

from geomanager.blocks import (
    InlineLegendBlock,
    LayerMoreInfoBlock,
    QueryParamSelectableBlock,
    QueryParamStaticBlock, InlineIconLegendBlock
)
from geomanager.models.core import BaseLayer
from geomanager.utils import DATE_FORMAT_CHOICES


class BaseTileLayer(TimeStampedModel, ClusterableModel, BaseLayer):
    class Meta:
        abstract = True

    base_url = models.CharField(max_length=500, verbose_name=_("Base Tile url"), )

    query_params_static = StreamField([
        ('param', QueryParamStaticBlock(label=_("Query Parameter")))
    ], use_json_field=True, null=True, blank=True, verbose_name=_("Static Query Params"),
        help_text=_("Static query params to be added to the url"))

    query_params_selectable = StreamField([
        ('param', QueryParamSelectableBlock(label=_("Selectable Query Parameter")))
    ], use_json_field=True, null=True, blank=True, verbose_name=_("Query Params With selectable Options"),
        help_text=_("This should provide a list of options that users "
                    "can choose to change the query parameter of the url"))

    params_selectors_side_by_side = models.BooleanField(default=False,
                                                        verbose_name=_("Arrange Param Selectors side by side"))

    legend = StreamField([
        ('legend', InlineLegendBlock(label=_("Custom Legend")),),
        ('legend_image', ImageChooserBlock(label=_("Custom Image")),),
        ('legend_icon', InlineIconLegendBlock(label=_("Legend Icon")),),
    ], use_json_field=True, null=True, blank=True, max_num=1, verbose_name=_("Legend"), )

    more_info = StreamField([
        ('more_info', LayerMoreInfoBlock(label=_("Info link")),),
    ], block_counts={
        'more_info': {'max_num': 1},
    }, use_json_field=True, null=True, blank=True, max_num=1, verbose_name=_("More Info"), )

    get_time_from_tile_json = models.BooleanField(default=False, verbose_name=_("Get time from tile json url"))
    tile_json_url = models.URLField(max_length=500, blank=True, null=True, verbose_name=_("Tile JSON url"))
    timestamps_response_object_key = models.CharField(max_length=100, blank=True, null=True, default="timestamps",
                                                      verbose_name=_("Timestamps response object key"),
                                                      help_text=_("Key for timestamps values in response object"))
    date_format = models.CharField(max_length=100, choices=DATE_FORMAT_CHOICES, blank=True, null=True,
                                   verbose_name=_("Display Format for DateTime Selector"))
    time_parameter_name = models.CharField(max_length=100, blank=True, null=True, default="time",
                                           verbose_name=_("Time Parameter Name"),
                                           help_text=_("Name of the time parameter in the url"))

    panels = [
        FieldPanel("title"),
        FieldPanel("default"),
        FieldPanel("base_url"),

        MultiFieldPanel([
            FieldPanel("get_time_from_tile_json"),
            FieldPanel("tile_json_url", classname="show_if_get_time_checked"),
            FieldPanel("timestamps_response_object_key", classname="show_if_get_time_checked"),
            FieldPanel("time_parameter_name", classname="show_if_get_time_checked"),
            FieldPanel("date_format", classname="show_if_get_time_checked"),
        ], heading=_("Time Settings")),

        FieldPanel("query_params_static"),
        FieldPanel("query_params_selectable"),
        FieldPanel("params_selectors_side_by_side"),
        FieldPanel("legend"),
        FieldPanel("more_info"),
    ]

    @property
    def tile_url(self):
        tile_url = self.base_url

        query_params = {}

        if self.has_time:
            time_param_name = self.time_parameter_name or "time"
            query_params.update({
                time_param_name: "{{time}}"
            })

        static_params = self.get_static_params()
        if static_params:
            query_params.update(static_params)

        query_str = '&'.join([f"{key}={value}" for key, value in query_params.items()])

        if query_str:
            tile_url = f"{tile_url}?{query_str}"

        return tile_url

    def get_selectable_params(self):
        params = {}
        if self.query_params_selectable:
            for query_param in self.query_params_selectable:
                data = query_param.block.get_api_representation(query_param.value)
                val = f"{data.get('name')}"
                params.update({val: data})
        return params

    def get_static_params(self):
        params = {}
        if self.query_params_static:
            for query_param in self.query_params_static:
                data = query_param.block.get_api_representation(query_param.value)
                val = f"{data.get('key')}"
                params.update({val: data.get('value')})
        return params

    def get_selectable_params_config(self):
        selectable_params = self.get_selectable_params()
        config = []

        for key, param_config in selectable_params.items():
            param_config = {
                "key": key,
                "required": True,
                "type": param_config.get("type"),
                "options": param_config.get("options"),
                "sentence": f"{param_config.get('label') or key} {{selector}}",
            }

            config.append(param_config)

        return config

    @property
    def has_time(self):
        return bool(self.dataset.multi_temporal and self.get_time_from_tile_json and self.tile_json_url)

    @property
    def params(self):
        params = {}
        if self.has_time:
            time_param_name = self.time_parameter_name or "time"
            params.update({time_param_name: ""})

        selector_config = self.get_selectable_params_config()

        for selector_param in selector_config:
            default = None
            for option in selector_param.get("options"):
                if option.get("default"):
                    default = option.get("value")
                    break
            if not default:
                default = selector_param.get("options")[0].get("value")
            params.update({selector_param.get("key"): default})

        return params

    @property
    def time_parameter_selector_config(self):
        time_param_name = self.time_parameter_name or "time"
        time_config = None
        if self.has_time:
            time_config = {
                "key": "time",
                "url_param": time_param_name,
                "required": True,
                "sentence": "{selector}",
                "type": "datetime",
                "availableDates": [],
            }

            if self.date_format:
                if self.date_format == "pentadal":
                    time_config.update({
                        "dateFormat": {"currentTime": "MMM yyyy", "asPeriod": "pentadal"},
                    })
                elif self.date_format == "dekadal":
                    time_config.update({
                        "dateFormat": {"currentTime": "MMM yyyy", "asPeriod": "dekadal"},
                    })
                else:
                    time_config.update({
                        "dateFormat": {"currentTime": self.date_format},
                    })
            else:
                time_config.update({
                    "dateFormat": {"currentTime": "yyyy-MM-dd HH:mm"},
                })

        return time_config

    @property
    def param_selector_config(self):
        config = []
        time_config = self.time_parameter_selector_config

        if time_config:
            config.append(time_config)

        selectable_params_config = self.get_selectable_params_config()
        config.extend(selectable_params_config)

        return config

    def get_legend_config(self, request):

        # default config
        config = {
            "type": "basic",
            "items": []
        }

        legend_block = self.legend

        # only one legend block entry is expected
        if legend_block:
            legend_block = legend_block[0]

        if legend_block:
            if isinstance(legend_block.value, Image):
                image_url = legend_block.value.file.url
                if request:
                    image_url = get_full_url(request, image_url)
                config.update({"type": "image", "imageUrl": image_url})
                return config

            data = legend_block.block.get_api_representation(legend_block.value)

            if legend_block.block_type == "legend_icon":
                for item in data.get("items", []):
                    config["items"].append({
                        "icon": item.get("icon_image"),
                        "name": item.get("icon_label"),
                        "color": item.get("icon_color"),
                        "iconSource": "sprite",
                    })
                return config

            config.update({"type": data.get("type")})

            for item in data.get("items"):
                config["items"].append({
                    "name": item.get("value"),
                    "color": item.get("color")
                })

        return config



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _
from django_extensions.db.models import TimeStampedModel
from modelcluster.fields import ParentalKey
from modelcluster.models import ClusterableModel
from wagtail.admin.panels import FieldPanel, FieldRowPanel, MultiFieldPanel, InlinePanel
from wagtail.models import Orderable
from wagtail_color_panel.edit_handlers import NativeColorPanel
from wagtail_color_panel.fields import ColorField

from geomanager.forms import RasterStyleModelForm
from geomanager.utils import significant_digits, round_to_precision
from geomanager.widgets import RasterStyleWidget


class RasterStyle(TimeStampedModel, ClusterableModel):
    LEGEND_TYPE_CHOICES = (
        ("basic", _("Basic")),
        ("choropleth", _("Choropleth Horizontal")),
        ("choropleth_vertical", _("Choropleth Vertical")),
        ("gradient", _("Gradient Horizontal")),
        ("gradient_vertical", _("Gradient Vertical")),
    )
    
    base_form_class = RasterStyleModelForm
    
    name = models.CharField(max_length=256, verbose_name=_("name"),
                            help_text=_("Style name for identification"))
    unit = models.CharField(max_length=100, blank=True, null=True, verbose_name=_("data unit"),
                            help_text=_("Data unit"))
    min = models.IntegerField(default=0, verbose_name=_("minimum value"), help_text=_("minimum value"))
    max = models.IntegerField(default=100, verbose_name=_("maximum value"), help_text=_("maximum value"))
    steps = models.IntegerField(default=5, validators=[MinValueValidator(3), MaxValueValidator(20), ], null=True,
                                blank=True, verbose_name=_("steps"), help_text=_("Number of steps"))
    use_custom_colors = models.BooleanField(default=False, verbose_name=_("Use Custom Colors"))
    palette = models.TextField(blank=True, null=True, verbose_name=_("Color Palette"))
    interpolate = models.BooleanField(default=False, verbose_name=_("interpolate"), help_text="Interpolate colorscale")
    legend_type = models.CharField(max_length=100, choices=LEGEND_TYPE_CHOICES, default="choropleth_vertical",
                                   verbose_name=_("Legend Type"))
    custom_color_for_rest = ColorField(blank=True, null=True, default="#ff0000",
                                       verbose_name=_("Color for the rest of values"),
                                       help_text=_(
                                           "Color for values greater than the values defined above, "
                                           "as well as values greater than the maximum defined value"))
    
    class Meta:
        verbose_name = _("Raster Style")
        verbose_name_plural = _("Raster Styles")
    
    def __str__(self):
        return self.name
    
    panels = [
        FieldPanel("name"),
        FieldPanel("unit"),
        FieldRowPanel(
            [
                FieldPanel("min"),
                FieldPanel("max"),
            ], _("Data values")
        ),
        FieldPanel("steps"),
        FieldPanel("palette", widget=RasterStyleWidget),
        FieldPanel("legend_type"),
        FieldPanel("use_custom_colors"),
        MultiFieldPanel([
            InlinePanel("color_values", heading=_("Color Values"), label=_("Color Value")),
            NativeColorPanel("custom_color_for_rest"),
        ], _("Custom Color Values")),
    ]
    
    def get_palette_list(self):
        if not self.use_custom_colors:
            return self.palette.split(",")
        return self.get_custom_palette()
    
    @property
    def min_value(self):
        return self.min
    
    @property
    def max_value(self):
        max_value = self.max
        if self.min == max_value:
            max_value += 0.1
        return max_value
    
    @property
    def scale_value(self):
        return 254 / (self.max_value - self.min_value)
    
    @property
    def offset_value(self):
        return -self.min_value
    
    @property
    def clip_value(self):
        return self.max_value + self.offset_value
    
    @property
    def palette_legend_values(self):
        if self.use_custom_colors:
            return None
        
        colors = self.get_palette_list()
        step = (self.max - self.min) / (len(colors) - (2 if self.min > 0 else 1))
        precision = significant_digits(step, self.max)
        value_format = round_to_precision(precision)
        
        val_from = self.min
        val_to = value_format(self.min + step)
        
        items = []
        
        for i, color in enumerate(colors):
            item = {"color": color}
            
            if i == 0 and self.min > 0:
                item.update({
                    "from": 0,
                    "to": self.min,
                    "name": "< {}".format(self.min)
                })
                val_to = self.min
            
            elif val_from < self.max:
                item.update({
                    "from": round(val_from, 1),
                    "to": round(val_to, 1),
                    "name": "{} - {}".format(val_from, val_to)
                })
            else:
                item.update({
                    "from": val_from,
                    "name": "> {}".format(val_from)
                })
            
            val_from = val_to
            val_to = value_format(self.min + step * (i + (1 if self.min > 0 else 2)))
            
            items.append(item)
        
        return items
    
    def get_custom_color_values(self):
        values = []
        color_values = self.color_values.order_by('threshold')
        
        for i, c_value in enumerate(color_values):
            value = c_value.value
            # if not the first one, add prev value for later comparison
            if i == 0:
                value["min_value"] = None
            else:
                value["min_value"] = color_values[i - 1].threshold
            value["max_value"] = value["threshold"]
            values.append(value)
        
        return values
    
    def get_custom_palette(self):
        colors = []
        for i in range(256):
            color = self.get_color_for_index(i)
            colors.append(color)
        
        return colors
    
    def get_color_for_index(self, index_value):
        values = self.get_custom_color_values()
        
        for value in values:
            max_value = value["max_value"] + self.offset_value
            
            if max_value > self.clip_value:
                max_value = self.clip_value
            
            if max_value < 0:
                max_value = 0
            
            max_value = self.scale_value * max_value
            
            if value["min_value"] is None:
                if index_value <= max_value:
                    return values[0]["color"]
            
            if value["min_value"] is not None:
                min_value = value["min_value"] + self.offset_value
                
                if min_value > self.clip_value:
                    min_value = self.clip_value
                
                if min_value < 0:
                    min_value = 0
                
                min_value = self.scale_value * min_value
                
                if min_value < index_value <= max_value:
                    return value["color"]
        
        return self.custom_color_for_rest
    
    def get_style_as_json(self):
        palette = self.get_palette_list()
        style = {
            "bands": [
                {
                    "band": 1,
                    "min": self.min,
                    "max": self.max,
                    "palette": palette,
                    "scheme": "discrete",
                }
            ]
        }
        return style
    
    def get_legend_config(self):
        items = []
        legend_type = self.legend_type
        if self.use_custom_colors:
            values = self.get_custom_color_values()
            count = len(values)
            
            if count > 0:
                for value in values:
                    item = {
                        "name": "",
                        "color": value['color']
                    }
                    if value.get("show_on_legend"):
                        item.update({
                            "name": value['label'] if value.get('label') else value['threshold'],
                        })
                    items.append(item)
                
                # if only one item and it is the custom color for rest, then show it as basic legend
                # no matter what the legend type was set
                if count == 1 and items[0].get("color") == self.custom_color_for_rest:
                    legend_type = "basic"
                else:
                    rest_item = {"name": "", "color": self.custom_color_for_rest}
                    items.append(rest_item)
        else:
            items = self.palette_legend_values
        
        config = {"type": legend_type, "items": items, }
        
        if self.unit:
            config["units"] = self.unit
        return config


class ColorValue(TimeStampedModel, Orderable):
    layer = ParentalKey(RasterStyle, related_name='color_values')
    threshold = models.FloatField(verbose_name=_("Threshold value"), help_text=_(
        "Values less than or equal to the input value, will be assigned the chosen color"))
    color = ColorField(default="#ff0000", verbose_name=_("color"))
    show_on_legend = models.BooleanField(default=True, verbose_name=_("Show label on Legend"))
    label = models.CharField(max_length=100, blank=True, null=True, verbose_name=_('Optional Label'))
    
    class Meta:
        verbose_name = _("Color Value")
        verbose_name_plural = _("Color Values")
    
    panels = [
        FieldPanel("threshold"),
        NativeColorPanel("color"),
        FieldPanel("show_on_legend"),
        FieldPanel("label")
    ]
    
    @property
    def value(self):
        return {
            "threshold": self.threshold,
            "color": self.color,
            "label": self.label,
            "show_on_legend": self.show_on_legend
        }



# Generated by Django 4.1.10 on 2023-11-06 09:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geomanager', '0028_alter_pgvectortable_options_alter_dataset_category'),
    ]

    operations = [
        migrations.AddField(
            model_name='rasterfilelayer',
            name='auto_ingest_from_directory',
            field=models.BooleanField(default=False, verbose_name='Auto ingest from directory'),
        ),
    ]
